/**         @file  memdbg.c
 *         @brief  Memory leak tracker implementation
 *          @date  25/09/2006
 *        @author  François Cerbelle (Fanfan), francois@cerbelle.net
 *     @copyright  Copyright (c) 1997-2024, François Cerbelle
 *
 *  Originally inspired by "L'art du code", Steve Maguire, Microsoft Press
 *
 *   @internal
 *       Compiler  gcc
 *  Last modified  2024-08-21 19:23
 *   Organization  Cerbelle.net
 *        Company  Home
 *
 *   This source code is released for free distribution under the terms of the
 *   GNU General Public License as published by the Free Software Foundation.
 */

#include "memdbg.h"
#include <string.h> /* memset, memcpy, memmove */
#include <stdio.h> /* asprintf */
#include <stdarg.h> /* va_list, va_start, va_arg, va_end */

/* Documented in header file */
void* dbg_malloc(
    const size_t Size,
    const char* File,
    const int Line,
    const char* CompilDate,
    const char* CompilTime,
    const char* Function
)
{
    /* Memory allocation */
    void* l_tmp = malloc(Size);
    if (NULL == l_tmp) return (l_tmp);

    /* If successful, track the memory block */
#pragma GCC diagnostic push                             /* save the actual diag context */
#pragma GCC diagnostic ignored "-Wmaybe-uninitialized"  /* locally disable maybe warnings */
    if (0!=memtrack_addblock( l_tmp, Size, File,Line,CompilDate,CompilTime,Function)) {
#pragma GCC diagnostic pop                              /* restore previous diag context */
        /* If tracking fails, the whole allocation fails */
        free(l_tmp);
        l_tmp = NULL;
    };

    return l_tmp;
}

/* Documented in header file */
void dbg_free(
    void* Ptr,
    const char* File,
    const int Line,
    const char* CompilDate,
    const char* CompilTime,
    const char* Function
)
{
    /* If the pointer was not NULL, it was tracked, remove it from the tracked
     * list. If it was not tracked, removing returns an error. Abort the
     * process as it should never have an untracked pointer. Either it was
     * allocated from a non instrumented binary, or it was allocated from
     * a non monitored function (see memory.h) or there is a bug in the
     * memory leak tracker. */
    if (NULL!=Ptr)
        if (0!=memtrack_delblock(Ptr,File,Line,CompilDate,CompilTime,Function))
            abort();

    /* If the pointer was NULL or tracked, forward it to the real free */
    free(Ptr);
}

/* Documented in header file */
void* dbg_calloc(
    const size_t NMemb,
    const size_t Size,
    const char* File,
    const int Line,
    const char* CompilDate,
    const char* CompilTime,
    const char* Function
)
{
    void* l_tmp;

    /* Use the dbg_malloc function to allocate the memory */
    l_tmp = dbg_malloc(
                NMemb*Size,
                File,Line,CompilDate,CompilTime,Function);

    /* Implement the calloc specific behavior compared to simple malloc:
     * it fills the allocated memory block with 0 */
    if (NULL != l_tmp)
        memset((char*)l_tmp, 0, NMemb*Size);

    return l_tmp;
}

/* Documented in header file */
void* dbg_realloc(
    void* Ptr,
    const size_t Size,
    const char* File,
    const int Line,
    const char* CompilDate,
    const char* CompilTime,
    const char* Function
)
{
    size_t l_oldsize; /**< Existing bloc size */
    char *newblk;                               /**< New block */

    /* NULL is not tracked but valid */
    if (NULL==Ptr) {
        l_oldsize=0;
    } else {
        /* Fetch existing block size */
        l_oldsize = memtrack_getblocksize(Ptr);

        /* This is probably a bug in the memory tracker.
         * It should not track zero sized blocks */
        if (0==l_oldsize)
            abort();
    }

    /* If new size is 0, then act as free, like realloc */
    if (0==Size) {
        dbg_free(Ptr,
                 File,Line,CompilDate,CompilTime,Function
                );
        return Ptr;
    }

    /* New sized block allocation to simulate the worst case scenario and
     * test a pointer change, a data loss (in case of shrink) */
    newblk=(char*)dbg_malloc(
               Size,
               File,Line,CompilDate,CompilTime,Function
           );
    /* The new block can fail */
    /* The real realloc function could succeed here, in case of inplace
     * shrink in an OOM situation. */
    if (NULL == newblk) return (newblk);

    /* Copy only the relevant data from old block to new block, loosing extra
     * data in case of shrink, and not initializing new data in case of
     * increase */
    memcpy(newblk,(char*)Ptr,(l_oldsize<Size?l_oldsize:Size));

    /* Free old block */
    dbg_free(
        Ptr,
        File,Line,CompilDate,CompilTime,Function
    );

    return (void*) newblk;
}


/* Documented in header file */
char* dbg_strdup(
    const char* Ptr,
    const char* File,
    const int Line,
    const char* CompilDate,
    const char* CompilTime,
    const char* Function
)
{
    char* l_newblk;                             /** Copy address */

    /* Use strdup to actually copy the string with its own return values and
     * abort (SIGSEGV in case of NULL) */
    l_newblk=NULL;
    l_newblk=strdup(Ptr);

    /* If the copy succeeded, try to track the memory allocation */
    if (NULL != l_newblk)
        if (0!=memtrack_addblock(
                l_newblk,
                strlen(l_newblk)+1,
                File,Line,CompilDate,CompilTime,Function
            )) {
            /* If tracking fails, the whole operation is reverted and fails */
            free(l_newblk);
            l_newblk=NULL;
        };

    return l_newblk;
}


/* Documented in header file */
int dbg_asprintf(char **p_Ptr,
                 const char* p_Format,
                 const char *File,
                 const int Line,
                 const char *CompilDate,
                 const char *CompilTime, const char *Function,
                 ...)
{
    int l_returncode;

    /* NULL is not allowed, where would we store the result, then ? */
    if(NULL==p_Ptr)
        abort();

    /* Limit the scope of the variadic manipulation variables */
    {
        va_list l_ap;
        va_start (l_ap, Function);
        /* Use the original vasprintf to build the formatted string */
        /** @todo Implement a vasprintf wrapping function to catch allocation
         * and use it here */
        l_returncode = vasprintf(p_Ptr, p_Format, l_ap);
        va_end(l_ap);
    }

    /* If formatting succeeded, try to track the memory allocation */
    if (-1 != l_returncode)
        if (1==memtrack_addblock(
                *p_Ptr,
                strlen (*p_Ptr)+1,
                File,Line,CompilDate,CompilTime,Function
            )) {
            /* If tracking fails, the whole operation is reverted and fails */
            free(*p_Ptr);
            l_returncode=-2;
        };

    return l_returncode;
}

