/**         @file  memtrack.c
 *         @brief  Memory block metadata tracking implementation
 *          @date  25/09/2006
 *        @author  François Cerbelle (Fanfan), francois@cerbelle.net
 *     @copyright  Copyright (c) 1997-2024, François Cerbelle
 *
 *  Originally inspired by "L'art du code", Steve Maguire, Microsoft Press
 *
 *   @internal
 *       Compiler  gcc
 *  Last modified  2024-08-21 13:20
 *   Organization  Cerbelle.net
 *        Company  Home
 *
 *   This source code is released for free distribution under the terms of the
 *   GNU General Public License as published by the Free Software Foundation.
 */

#include "memtrack.h"
#include <string.h>             /* strdup */
#include <stdio.h>              /* fprintf */

static struct MemBlock *Head;   /**< Memory blocks metadata list start sentinal */
static struct MemBlock *Tail;   /**< Memory blocks metadata list end sentinal */
static uint64_t NbBlocks;       /**< Memory blocks metadata counter */
static uint64_t RAMSize;        /**< Memory blocks metadata RAM sum */

/* Early definition, documented in the implementation below */
static unsigned int memtrack_init();

/* Early definition, documented in the implementation below */
static unsigned int memtrack_addblock_preinit(const void *p_Ptr,
        const size_t p_Size,
        const char *p_File,
        const int p_Line,
        const char *p_CompilDate,
        const char *p_CompilTime,
        const char *p_Function);

/* Early definition, documented in the implementation below */
static unsigned int memtrack_delblock_preinit(const void *p_Ptr,
        const char *p_File,
        const int p_Line,
        const char *p_CompilDate,
        const char *p_CompilTime,
        const char *p_Function);

/* Early definition, documented in the implementation below */
static uint64_t memtrack_dumpblocks_preinit();

/* Early definition, documented in the implementation below */
static uint64_t memtrack_getallocatedblocks_preinit();

/* Early definition, documented in the implementation below */
static uint64_t memtrack_getallocatedRAM_preinit();

/* Early definition, documented in the implementation below */
static size_t memtrack_getblocksize_preinit(const void *p_Ptr);

/* Early definition, documented in the implementation below */
static unsigned int memtrack_addblock_postinit(const void *p_Ptr,
        const size_t p_Size,
        const char *p_File,
        const int p_Line,
        const char *p_CompilDate,
        const char *p_CompilTime,
        const char *p_Function);

/* Early definition, documented in the implementation below */
static unsigned int memtrack_delblock_postinit(const void *p_Ptr,
        const char *p_File,
        const int p_Line,
        const char *p_CompilDate,
        const char *p_CompilTime,
        const char *p_Function);

/* Early definition, documented in the implementation below */
static uint64_t memtrack_dumpblocks_postinit();

/* Early definition, documented in the implementation below */
static uint64_t memtrack_getallocatedblocks_postinit();

/* Early definition, documented in the implementation below */
static uint64_t memtrack_getallocatedRAM_postinit();

/* Early definition, documented in the implementation below */
static size_t memtrack_getblocksize_postinit(const void *p_Ptr);

/*******************/
/* Implementations */
/*******************/

/** Memory block metadata list reset
 *
 * This function is not defined in the header file because it should never be
 * used. It is exported because it is used by the unit tests. NEVER USE IT, it
 * will crash your code.
 */
void memtrack_reset()
{
    /* Ne réinitialise que si nécessaire */
    if ( memtrack_addblock == memtrack_addblock_preinit)
        return;

    /* Déconfiguration des foncteurs */
    memtrack_addblock = memtrack_addblock_preinit;
    memtrack_delblock = memtrack_delblock_preinit;
    memtrack_dumpblocks = memtrack_dumpblocks_preinit;
    memtrack_getallocatedblocks = memtrack_getallocatedblocks_preinit;
    memtrack_getallocatedRAM = memtrack_getallocatedRAM_preinit;
    memtrack_getblocksize = memtrack_getblocksize_preinit;

    /* Purge de la liste */
    while (Tail!=Head->Next) {
        memtrack_delblock_postinit(Head->Next->Ptr,NULL,0,NULL,NULL,NULL);
    }

    /* Réinitialization des pointeurs */
    free(Head);
    free(Tail);
    Head=NULL;
    Tail=NULL;
}

/** Memory blocks metadata list initialization
 *
 * This function is private, not meant to be called externally, it is
 * automatically called by other functions at the very first usage. It
 * initializes the memory block metadata list, with start and end sentinels. It
 * needs to be executed before the real list manipulation functions.
 *
 * Each real manipulation function has two private static implementations,
 * uninitialized and initialized, and a publicly exported functor which is
 * referencing the uninitialized form by default. The uninitialized forms first
 * call this initialization function which makes the functos reference the
 * initialized forms.
 *
 *  \return Initialization status
 *      \retval 0 in case of success
 */
static unsigned int memtrack_init()
{
    Head = (TMemBlock *) malloc(sizeof(TMemBlock));
    Tail = (TMemBlock *) malloc(sizeof(TMemBlock));

    if ((NULL==Head)||(NULL==Tail)) {
        fprintf(stderr,"%s:%d Not enough memory to initialize memtracker\n",
                __FILE__,__LINE__);
        return 1;
    }
    Head->Prev = (TMemBlock *) NULL;
    Head->Next = Tail;
    Tail->Next = (TMemBlock *) NULL;
    Tail->Prev = Head;
    Tail->Ptr = Head->Ptr = (void *)NULL;
    Tail->Size = Head->Size = 0;
    Tail->File = Head->File = (char *)NULL;
    Tail->Line = Head->Line = 0;
    Tail->CompilDate = Head->CompilDate = (char *)NULL;
    Tail->CompilTime = Head->CompilTime = (char *)NULL;
    Tail->Function = Head->Function = (char *)NULL;

    /* Initialisation des compteurs */
    NbBlocks=0;
    RAMSize=0;

    /* Modification des foncteurs pour utiliser les fonctions définitives */
    memtrack_addblock = memtrack_addblock_postinit;
    memtrack_delblock = memtrack_delblock_postinit;
    memtrack_dumpblocks = memtrack_dumpblocks_postinit;
    memtrack_getallocatedblocks = memtrack_getallocatedblocks_postinit;
    memtrack_getallocatedRAM = memtrack_getallocatedRAM_postinit;
    memtrack_getblocksize = memtrack_getblocksize_postinit;

    return 0;
}


/** Add a memory block metadata entry to the list (preinit)
 *
 * Default function for the memtrack_addblock functor until initialized by the
 * memtrack_init function. It is executed if the functor is called before
 * initialization. It triggers initialization with functors definitive
 * assignation to the *_postinit functions and calls the actual postinit
 * function.
 *
 * \sa memtrack_addblock
 * \sa memtrack_addblock_postinit
 */
static unsigned int memtrack_addblock_preinit(const void *p_Ptr,
        const size_t p_Size,
        const char *p_File,
        const int p_Line,
        const char *p_CompilDate,
        const char *p_CompilTime,
        const char *p_Function)
{
    /* Initialisation de la liste */
    if (0!=memtrack_init()) {
        fprintf(stderr,"%s:%d Not enough memory to initialize memtracker\n",
                __FILE__,__LINE__);
        /* OOM */
        return 1;
    }

    /* Appel de la fonction réelle */
    return memtrack_addblock(p_Ptr, p_Size, p_File, p_Line, p_CompilDate,
                             p_CompilTime, p_Function);
}

/** Remove a memory block metadata entry from the list (preinit)
 *
 * Default function for the memtrack_delblock functor until initialized by the
 * memtrack_init function. It is executed if the functor is called before
 * initialization. It triggers initialization with functors definitive
 * assignation to the *_postinit functions and calls the actual postinit
 * function.
 *
 * \sa memtrack_delblock
 * \sa memtrack_delblock_postinit
 */
static unsigned int memtrack_delblock_preinit(const void *p_Ptr,
        const char *p_File,
        const int p_Line,
        const char *p_CompilDate,
        const char *p_CompilTime,
        const char *p_Function)
{
    /* Initialisation de la liste */
    if (0!=memtrack_init()) {
        fprintf(stderr,"%s:%d Not enough memory to initialize memtracker\n",
                __FILE__,__LINE__);
        /* OOM */
        return 1;
    }

    /* Appel de la fonction réelle */
    return memtrack_delblock(p_Ptr, p_File, p_Line, p_CompilDate,
                             p_CompilTime, p_Function);
}

/** Dump memory block metadata entries from the list (preinit)
 *
 * Default function for the memtrack_dumpblocks functor until initialized by the
 * memtrack_init function. It is executed if the functor is called before
 * initialization. It triggers initialization with functors definitive
 * assignation to the *_postinit functions and calls the actual postinit
 * function.
 *
 * \sa memtrack_dumpblocks
 * \sa memtrack_dumpblocks_postinit
 */
static uint64_t memtrack_dumpblocks_preinit()
{
    /* Initialisation de la liste */
    if (0!=memtrack_init()) {
        fprintf(stderr,"%s:%d Not enough memory to initialize memtracker\n",
                __FILE__,__LINE__);
        /* OOM */
        return 0;
    }

    /* Appel de la fonction réelle */
    return memtrack_dumpblocks();
}

/** Get the memory block metadata entries count (preinit)
 *
 * Default function for the memtrack_getallocatedblocks functor until
 * initialized by the memtrack_init function. It is executed if the functor is
 * called before initialization. It triggers initialization with functors
 * definitive assignation to the *_postinit functions and calls the actual
 * postinit function.
 *
 * \sa memtrack_getallocatedblocks
 * \sa memtrack_getallocatedblocks_postinit
 */
static uint64_t memtrack_getallocatedblocks_preinit()
{
    /* Initialisation de la liste */
    if (0!=memtrack_init()) {
        fprintf(stderr,"%s:%d Not enough memory to initialize memtracker\n",
                __FILE__,__LINE__);
        /* OOM */
        return 0;
    }

    /* Appel de la fonction réelle */
    return memtrack_getallocatedblocks();
}

/** Get the memory block metadata entries total RAM (preinit)
 *
 * Default function for the memtrack_getallocatedRAM functor until
 * initialized by the memtrack_init function. It is executed if the functor is
 * called before initialization. It triggers initialization with functors
 * definitive assignation to the *_postinit functions and calls the actual
 * postinit function.
 *
 * \sa memtrack_getallocatedRAM
 * \sa memtrack_getallocatedRAM_postinit
 */
static uint64_t memtrack_getallocatedRAM_preinit()
{
    /* Initialisation de la liste */
    if (0!=memtrack_init()) {
        fprintf(stderr,"%s:%d Not enough memory to initialize memtracker\n",
                __FILE__,__LINE__);
        /* OOM */
        return 0;
    }

    /* Appel de la fonction réelle */
    return memtrack_getallocatedRAM();
}

/** Get a memory block RAM size (preinit)
 *
 * Default function for the memtrack_getblocksize functor until initialized by
 * the memtrack_init function. It is executed if the functor is called before
 * initialization. It triggers initialization with functors definitive
 * assignation to the *_postinit functions and calls the actual postinit
 * function.
 *
 * \sa memtrack_getblocksize
 * \sa memtrack_getblocksize_postinit
 */
static size_t memtrack_getblocksize_preinit(const void *p_Ptr)
{
    /* Initialisation de la liste */
    if (0!=memtrack_init()) {
        fprintf(stderr,"%s:%d Not enough memory to initialize memtracker\n",
                __FILE__,__LINE__);
        /* OOM */
        return 0;
    }

    /* Appel de la fonction réelle */
    return memtrack_getblocksize(p_Ptr);
}

/** Add a memory block metadata entry to the list (postinit)
 *
 * This function is referenced by the memtrack_addblock functor after
 * initialization. This is the real implementation of the feature.
 *
 * \sa memtrack_addblock
 * \sa memtrack_addblock_preinit
 */
static unsigned int memtrack_addblock_postinit(const void *p_Ptr,
        const size_t p_Size,
        const char *p_File,
        const int p_Line,
        const char *p_CompilDate,
        const char *p_CompilTime,
        const char *p_Function)
{
    TMemBlock *l_tmp;

    /* Test de validité des données à enregistrer */
    /* Meme si malloc permet Size=0, ce n'est pas portable */
    if ((NULL==p_Ptr)||
        (0==p_Size)||
        (NULL==p_File)||
        (0==p_File[0])||
        (0==p_Line)||
        (NULL==p_CompilDate)||
        (0==p_CompilDate[0])||
        (NULL==p_CompilTime)||
        (0==p_CompilTime[0])||
        (NULL==p_Function)||
        (0==p_Function[0])) {
        fprintf(stderr,"%s:%d Null or empty parameters\n",__FILE__, __LINE__);
        return 1;
    }

    /* On ne peut pas dupliquer un pointeur. Pour le modifier, il faut le
     * supprimer et le recréer, ce n'est pas le rôle de ces fonctions  de
     * bas-niveau */

    /* Recherche du pointeur */
    l_tmp = Head->Next;
    while ((l_tmp->Ptr != p_Ptr) && (l_tmp != Tail))
        l_tmp = l_tmp->Next;

    /* Le bloc ne doit pas avoir été trouvé */
    if (l_tmp != Tail) {
        fprintf(stderr,"%s:%d Memory bloc already registered\n",__FILE__, __LINE__);
        return 1;
    }

    /* Allocation d'un nouveau descripteur de bloc */
    l_tmp = (TMemBlock *) malloc(sizeof(TMemBlock));

    /* Allocation réussie ? */
    if (NULL == l_tmp) {
        return 1;
    }

    /* Remplissage du descripteur */
    l_tmp->Ptr = (void *)p_Ptr;
    l_tmp->Size = p_Size;
    if (NULL==(l_tmp->File = strdup(p_File?p_File:""))) {
        free(l_tmp);
        return 1;
    };
    l_tmp->Line = p_Line;
    if (NULL==(l_tmp->CompilDate = strdup(p_CompilDate?p_CompilDate:""))) {
        free(l_tmp->File);
        free(l_tmp);
        return 1;
    };
    if (NULL==(l_tmp->CompilTime = strdup(p_CompilTime?p_CompilTime:""))) {
        free(l_tmp->CompilDate);
        free(l_tmp->File);
        free(l_tmp);
        return 1;
    };
    if (NULL==(l_tmp->Function = strdup(p_Function?p_Function:""))) {
        free(l_tmp->CompilTime);
        free(l_tmp->CompilDate);
        free(l_tmp->File);
        free(l_tmp);
        return 1;
    };

    /* Ajout de la description dans la liste (Section critique) */
    l_tmp->Prev = Tail->Prev;
    l_tmp->Next = Tail;
    l_tmp->Prev->Next = l_tmp->Next->Prev = l_tmp;

    /* Mise à jour des compteurs */
    NbBlocks++;
    RAMSize += p_Size;

    return 0;
}

/** Remove a memory block metadata entry from the list (postinit)
 *
 * This function is referenced by the memtrack_addblock functor after
 * initialization. This is the real implementation of the feature.
 *
 * \sa memtrack_delblock
 * \sa memtrack_delblock_preinit
 */
static unsigned int memtrack_delblock_postinit(const void *p_Ptr,
        const char *p_File,
        const int p_Line,
        const char *p_CompilDate,
        const char *p_CompilTime,
        const char *p_Function)
{
    TMemBlock *l_tmp;
    (void) p_File;
    (void) p_Line;
    (void) p_CompilDate;
    (void) p_CompilTime;
    (void) p_Function;

    /* Recherche de la description */
    l_tmp = Head->Next;
    while ((l_tmp->Ptr != p_Ptr) && (l_tmp != Tail))
        l_tmp = l_tmp->Next;

    /* Le bloc doit avoir été trouvé */
    if (l_tmp == Tail) {
        fprintf(stderr,"%s:%d Block not found for deletion\n",__FILE__, __LINE__);
        return 1;
    }

    /* Libération des ressources acquises */
    /* On ne libère pas le bloc mémoire lui-même */
    free(l_tmp->File);
    free(l_tmp->CompilDate);
    free(l_tmp->CompilTime);
    free(l_tmp->Function);

    /* Retrait de la description de la liste (Section critique) */
    l_tmp->Next->Prev = l_tmp->Prev;
    l_tmp->Prev->Next = l_tmp->Next;

    /* Mise à jour des compteurs */
    NbBlocks--;
    RAMSize -= l_tmp->Size;

    /* Libération de la description */
    free(l_tmp);

    return 0;
}

/** Dump memory block metadata entries from the list (postinit)
 *
 * This function is referenced by the memtrack_addblock functor after
 * initialization. This is the real implementation of the feature.
 *
 * \sa memtrack_dumpblocks
 * \sa memtrack_dumpblocks_preinit
 */
static uint64_t memtrack_dumpblocks_postinit()
{
    TMemBlock *l_tmp;
    uint64_t l_NbBlocks = 0;;

    if (Head->Next != Tail) {
        size_t l_BlockSize;
        fprintf(stderr,
                "+----------------------------------------------------------------------------------------------------------+\n");
        fprintf(stderr, "| %-104s |\n", "Memory Tracker Report");
        fprintf(stderr,
                "+----------------------+----------------------+------+-----------------+----------+------------------------+\n");
        fprintf(stderr,
                "| %-20s | %-20s | %-4s | %-15s | %-8s | %-22s |\n",
                "Function", "File", "Line", "Address", "Bytes",
                "Compiled");
        fprintf(stderr,
                "+----------------------+----------------------+------+-----------------+----------+------------------------+\n");

        l_tmp = Head->Next;  /**< Curseur */
        l_BlockSize = 0;
        while (l_tmp != Tail) {
            fprintf(stderr,
                    "| %-20s | %-20s | %4d | %15p | %8lu | %11s @ %8s |\n",
                    l_tmp->Function, l_tmp->File, l_tmp->Line,
                    l_tmp->Ptr, (unsigned long)l_tmp->Size, l_tmp->CompilDate,
                    l_tmp->CompilTime);
            l_NbBlocks++;
            l_BlockSize += l_tmp->Size;
            l_tmp = l_tmp->Next;
        }
        fprintf(stderr,
                "+----------------------+----------------------+------+-----------------+----------+------------------------+\n");
        fprintf(stderr,
                "| %9lu bytes in %6lu blocks. %70s |\n", (unsigned long)l_BlockSize,
                (unsigned long)l_NbBlocks, "");
        fprintf(stderr,
                "+----------------------------------------------------------------------------------------------------------+\n");
    }
    /* Add ASSERT or assert (l_NbBlocks==NbBlocks) */
    return l_NbBlocks;
}

/** Get the memory block metadata entries count (postinit)
 *
 * This function is referenced by the memtrack_addblock functor after
 * initialization. This is the real implementation of the feature.
 *
 * \sa memtrack_getallocatedblocks
 * \sa memtrack_getallocatedblocks_preinit
 */
static uint64_t memtrack_getallocatedblocks_postinit()
{
    return NbBlocks;
}

/** Get the memory block metadata entries total RAM (postinit)
 *
 * This function is referenced by the memtrack_addblock functor after
 * initialization. This is the real implementation of the feature.
 *
 * \sa memtrack_getallocatedRAM
 * \sa memtrack_getallocatedRAM_preinit
 */
static uint64_t memtrack_getallocatedRAM_postinit()
{
    return RAMSize;
}

/** Get a memory block RAM size (postinit)
 *
 * This function is referenced by the memtrack_addblock functor after
 * initialization. This is the real implementation of the feature.
 *
 * \sa memtrack_getblocksize
 * \sa memtrack_getblocksize_preinit
 */
static size_t memtrack_getblocksize_postinit(const void *p_Ptr)
{
    /* Recherche de la description */
    TMemBlock *l_tmp = Head->Next;
    while ((l_tmp->Ptr != p_Ptr) && (l_tmp != Tail))
        l_tmp = l_tmp->Next;

    /* Le bloc doit avoir été trouvé */
    if (l_tmp == Tail)
        return 0;
    else
        return l_tmp->Size;
}

/* Documented in header file */
unsigned int (*memtrack_addblock) (const void *p_Ptr,
                                   const size_t p_Size,
                                   const char *p_File,
                                   const int p_Line,
                                   const char *p_CompilDate,
                                   const char *p_CompilTime,
                                   const char *p_Function) =
                                       memtrack_addblock_preinit;

/* Documented in header file */
unsigned int (*memtrack_delblock) (const void *p_Ptr,
                                   const char *p_File,
                                   const int p_Line,
                                   const char *p_CompilDate,
                                   const char *p_CompilTime,
                                   const char *p_Function) =
                                       memtrack_delblock_preinit;

/* Documented in header file */
uint64_t (*memtrack_dumpblocks) () = memtrack_dumpblocks_preinit;

/* Documented in header file */
uint64_t (*memtrack_getallocatedblocks) () = memtrack_getallocatedblocks_preinit;

/* Documented in header file */
uint64_t (*memtrack_getallocatedRAM) () = memtrack_getallocatedRAM_preinit;

/* Documented in header file */
size_t(*memtrack_getblocksize) (const void *p_Ptr) = memtrack_getblocksize_preinit;
